/* IniReader.c
 * $Format: " * Version $ProjectVersion$"$
 * Version mm7323.25
 * Michael McTernan, mm7323@bris.ac.uk
 * 17/3/00
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "IniReader.h"
#include "Set.h"

#define Error(a)    {fprintf(stderr,a);abort();}
#define Warning(a)  {fprintf(stderr,a);}


/* Static functions */
static int  iscomment(char c);
static int  ReadLine(char *Dest,FILE *f);
static long HashKey(char *Line);
static void ParseLine(char *Line);

static set_t Table;

/*INI_Init*/
/* Creates the hash table for fast lookups 
 */
static void INI_Init(ini_t *Keys) {
  int t=0,u;
  ini_t *e;

  /* Count number of elements and create the hashkeys */
  while(Keys[t].Key!=NULL) {
    Keys[t].HashKey=HashKey(Keys[t].Key);
    t++;
  }

  /* Create the hash table */
  Table=SET_New(t,sizeof(long));

  /* Add the data */
  for(u=0;u<t;u++)
  {
    if((e=SET_Get(Table,&Keys[u].HashKey))!=NULL)
    {
      char charbuf[50];
      sprintf(charbuf, "Non unique hash code for %s and %s",Keys[u].Key, e->Key);
      Warning(charbuf);
    }

    SET_Add(Table,&Keys[u],&Keys[u].HashKey);
  }
}


/*INI_Destroy*/
/* Deletes the hash table
 */
static void INI_Destroy() {
  SET_Destroy(Table);
}

/*INI_ReadFile*/
/* Read the file and get all options
 */
void INI_ReadFile(ini_t *Keys,char *Filename) {
  char Buffer[1024];
  FILE *File;

  INI_Init(Keys);

  if((File=fopen(Filename,"r"))==NULL)
    Error("fopen failed");

  /* Process each line */
  while(ReadLine(Buffer,File)>0) {
  
    /* Skip comments */
    if(!iscomment(*Buffer))
      ParseLine(Buffer);

  }

  INI_Destroy();
}


/*ParseLine*/
/* Parse a single line
 */
static void ParseLine(char *Line) {
  char  *Key,*Value;
  long   HashValue;
  ini_t *Entry;  
  int t=0;

  /* Find the = sign */
  while(t<strlen(Line) && Line[t]!='=')
    t++;

  /* Isolate the left and right side of equal sign */
  Key=Line;
  Line[t]='\0';
  Value=&Line[t+1];

  /* Calculate the hash value and find the entry */
  HashValue=HashKey(Key);
  Entry=SET_Get(Table,&HashValue);

  if(Entry==NULL) {
    fprintf(stderr,"%s=%s\n",Key,Value);
    Warning("Line not recognised");
    return;
  }

  /* Put in the data */
  if(Entry->Array==NULL)
    t=sscanf(Value,Entry->Format,Entry->Data);  
  else {
    t=sscanf(Value,Entry->Format,
        &((char *)Entry->Data)[*Entry->Array*Entry->ElemSize]);  
  }

  /* Check the return value of scanf */
  if(t<1) {
    fprintf(stderr,"%s=%s\n",Key,Value);
    fprintf(stderr,"Expecting %s\n",Entry->Format);
    fprintf(stderr,"%d\n",t);
    Warning("Key value failed to parse\n");
  }

}


/*iscomment*/
/* Return if c is a comment character 
 */
static int iscomment(char c) {
  return c=='#' || c==';' || c=='/';
}


/*ReadLine*/
/* Read a line and put it in the supplied
 * buffer.
 */
static int ReadLine(char *Dest,FILE *f) {
  int t=0;
  char c;

  do {
    c=fgetc(f);
    if(c!='\n') Dest[t++]=c;
  } while(c!='\n' && c!=EOF);

  Dest[t++]='\0';  

  /* Skip blank lines */
  if(t==1)
    return ReadLine(Dest,f);
  else
    return c==EOF? -1:t-1;
}


/*HashKey*/
/* Creates a hash key, ignoring spaces and case
 */
static long HashKey(char *Line) {
  int t;
  long Total=0;

  for(t=0;t<strlen(Line);t++) {
    if(!isspace((int)Line[t]))
      Total+=tolower(Line[t])*t;
  }

  Total+=strlen(Line);

  return Total;
}
